@@ -24,15 +24,14 @@
 
 package net.sf.picard.analysis;
 
+import net.sf.picard.reference.ReferenceSequence;
+import net.sf.picard.util.CollectionUtil;
 import net.sf.picard.util.Histogram;
 import net.sf.picard.sam.ReservedTagConstants;
-import net.sf.picard.cmdline.CommandLineProgram;
 import net.sf.picard.cmdline.Option;
 import net.sf.picard.cmdline.Usage;
-import net.sf.picard.cmdline.StandardOptionDefinitions;
 import net.sf.picard.io.IoUtil;
 import net.sf.picard.metrics.*;
-import net.sf.picard.reference.ReferenceSequenceFileWalker;
 import net.sf.picard.analysis.AlignmentSummaryMetrics.Category;
 import net.sf.picard.util.IlluminaUtil;
 import net.sf.picard.util.Log;
@@ -40,7 +39,6 @@ import net.sf.samtools.util.CoordMath;
 import net.sf.samtools.util.SequenceUtil;
 import net.sf.samtools.AlignmentBlock;
 import net.sf.samtools.SAMFileHeader;
-import net.sf.samtools.SAMFileReader;
 import net.sf.samtools.SAMRecord;
 import net.sf.samtools.util.StringUtil;
 
@@ -68,7 +66,7 @@ import java.util.*;
  * 
  * @author Doug Voet (dvoet at broadinstitute dot org)
  */
-public class CollectAlignmentSummaryMetrics extends CommandLineProgram {
+public class CollectAlignmentSummaryMetrics extends SinglePassSamProgram {
     private static final int MAPPING_QUALITY_THRESHOLD = 20;
     private static final int BASE_QUALITY_THRESHOLD = 20;
 
@@ -77,83 +75,70 @@ public class CollectAlignmentSummaryMetrics extends CommandLineProgram {
     private byte[][] ADAPTER_SEQUENCES;
     private static final Log log = Log.getInstance(CollectAlignmentSummaryMetrics.class);
 
-
+    final AlignmentSummaryMetricsCollector unpairedCollector     = new AlignmentSummaryMetricsCollector(Category.UNPAIRED);
+    final AlignmentSummaryMetricsCollector firstOfPairCollector  = new AlignmentSummaryMetricsCollector(Category.FIRST_OF_PAIR);
+    final AlignmentSummaryMetricsCollector secondOfPairCollector = new AlignmentSummaryMetricsCollector(Category.SECOND_OF_PAIR);
+    final AlignmentSummaryMetricsCollector pairCollector         = new AlignmentSummaryMetricsCollector(Category.PAIR);
 
     // Usage and parameters
     @Usage
     public String USAGE = "Reads a SAM or BAM file and writes a file containing summary alignment metrics.\n";
-    @Option(shortName="I", doc="SAM or BAM file") public File INPUT;
-    @Option(shortName="O", doc="File to write insert size metrics to") public File OUTPUT;
-    @Option(shortName="R", doc="Reference sequence file") public File REFERENCE_SEQUENCE;
-    @Option(doc="If true (default), \"unsorted\" SAM/BAM files will be considerd coordinate sorted",
-            shortName = StandardOptionDefinitions.ASSUME_SORTED_SHORT_NAME)
-    public Boolean ASSUME_SORTED = Boolean.TRUE;
+
     @Option(doc="Paired end reads above this insert size will be considered chimeric along with inter-chromosomal pairs.")
     public int MAX_INSERT_SIZE = 100000;
-    @Option() public List<String> ADAPTER_SEQUENCE = Arrays.asList(
-        IlluminaUtil.AdapterPair.SINGLE_END.get5PrimeAdapter(),
-        IlluminaUtil.AdapterPair.SINGLE_END.get3PrimeAdapter(),
-        IlluminaUtil.AdapterPair.PAIRED_END.get5PrimeAdapter(),
-        IlluminaUtil.AdapterPair.PAIRED_END.get3PrimeAdapter(),
-        IlluminaUtil.AdapterPair.INDEXED.get5PrimeAdapter(),
-        IlluminaUtil.AdapterPair.INDEXED.get3PrimeAdapter()
-    );  // todo - this list is immutable, so this option doesn't really work
-    // also List options are only appended to, not replaced
+
+    @Option() public List<String> ADAPTER_SEQUENCE = CollectionUtil.makeList(
+        IlluminaUtil.IlluminaAdapterPair.SINGLE_END.get5PrimeAdapter(),
+        IlluminaUtil.IlluminaAdapterPair.SINGLE_END.get3PrimeAdapter(),
+        IlluminaUtil.IlluminaAdapterPair.PAIRED_END.get5PrimeAdapter(),
+        IlluminaUtil.IlluminaAdapterPair.PAIRED_END.get3PrimeAdapter(),
+        IlluminaUtil.IlluminaAdapterPair.INDEXED.get5PrimeAdapter(),
+        IlluminaUtil.IlluminaAdapterPair.INDEXED.get3PrimeAdapter()
+    );
+
     @Option(shortName="BS", doc="Whether the SAM or BAM file consists of bisulfite sequenced reads.  ")
     public boolean IS_BISULFITE_SEQUENCED = false;
 
-    private ReferenceSequenceFileWalker referenceSequenceWalker;
-    private SAMFileHeader samFileHeader;
+    private boolean doRefMetrics;
 
     /** Required main method implementation. */
     public static void main(final String[] argv) {
-        System.exit(new CollectAlignmentSummaryMetrics().instanceMain(argv));
+        new CollectAlignmentSummaryMetrics().instanceMainWithExit(argv);
     }
 
-    @Override
-    protected int doWork() {
+    /** Silly method that is necessary to give unit test access to call doWork() */
+    protected final int testDoWork() { return doWork(); }
+
+    @Override protected void setup(final SAMFileHeader header, final File samFile) {
         prepareAdapterSequences();
-
-        // Check the files are readable/writable
-        IoUtil.assertFileIsReadable(INPUT);
-        IoUtil.assertFileIsReadable(REFERENCE_SEQUENCE);
+        doRefMetrics = REFERENCE_SEQUENCE != null;
         IoUtil.assertFileIsWritable(OUTPUT);
-        final SAMFileReader in = new SAMFileReader(INPUT);
-        assertCoordinateSortOrder(in);
 
-        this.referenceSequenceWalker = new ReferenceSequenceFileWalker(REFERENCE_SEQUENCE);
-        this.samFileHeader = in.getFileHeader();
-
-        if (!samFileHeader.getSequenceDictionary().isEmpty()) {
-            SequenceUtil.assertSequenceDictionariesEqual(
-                    samFileHeader.getSequenceDictionary(),
-                    this.referenceSequenceWalker.getSequenceDictionary());
-        } else {
+        if (header.getSequenceDictionary().isEmpty()) {
             log.warn(INPUT.getAbsoluteFile() + " has no sequence dictionary.  If any reads " +
                     "in the file are aligned then alignment summary metrics collection will fail.");
         }
 
-        final MetricCollector<AlignmentSummaryMetrics, SAMRecord> unpairedCollector =  constructCollector(Category.UNPAIRED);
-        final MetricCollector<AlignmentSummaryMetrics, SAMRecord> firstOfPairCollector = constructCollector(Category.FIRST_OF_PAIR);
-        final MetricCollector<AlignmentSummaryMetrics, SAMRecord> secondOfPairCollector = constructCollector(Category.SECOND_OF_PAIR);
-        final MetricCollector<AlignmentSummaryMetrics, SAMRecord> pairCollector =  constructCollector(Category.PAIR);
+    }
 
-        // Loop over the reads applying them to the correct collectors
-        for (SAMRecord record : in) {
-            if (record.getReadPairedFlag()) {
-                if (record.getFirstOfPairFlag()) firstOfPairCollector.addRecord(record);
-                else secondOfPairCollector.addRecord(record);
-
-                pairCollector.addRecord(record);
+    @Override protected void acceptRead(final SAMRecord rec, final ReferenceSequence ref) {
+        if (rec.getNotPrimaryAlignmentFlag()) return;
+        if (rec.getReadPairedFlag()) {
+            if (rec.getFirstOfPairFlag()) {
+                firstOfPairCollector.addRecord(rec, ref);
             }
             else {
-                unpairedCollector.addRecord(record);
+                secondOfPairCollector.addRecord(rec, ref);
             }
 
+            pairCollector.addRecord(rec, ref);
         }
+        else {
+            unpairedCollector.addRecord(rec, ref);
+        }
+    }
 
-        in.close();
-
+    @Override protected void finish() {
         // Let the collectors do any summary computations etc.
         firstOfPairCollector.onComplete();
         secondOfPairCollector.onComplete();
@@ -173,32 +158,12 @@ public class CollectAlignmentSummaryMetrics extends CommandLineProgram {
             file.addMetric(pairCollector.getMetrics());
         }
 
-        if (unpairedCollector.getMetrics().TOTAL_READS > 0) {
+        //if there are no reads in any category then we will returned an unpaired alignment summary metric with all zero values
+        if (unpairedCollector.getMetrics().TOTAL_READS > 0 || firstOfPairCollector.getMetrics().TOTAL_READS == 0) {
             file.addMetric(unpairedCollector.getMetrics());
         }
 
         file.write(OUTPUT);
-
-        return 0;
-    }
-
-    /**
-     * Checks that the SAM is either coordinate sorted according to it's header, or that the header
-     * doesn't specify a sort and that the user has supplied the argument to assume that it is
-     * sorted correctly.
-     */
-    private void assertCoordinateSortOrder(final SAMFileReader in) {
-        switch (in.getFileHeader().getSortOrder()) {
-        case coordinate:
-            break;
-        case unsorted:
-            if (this.ASSUME_SORTED) {
-                break;
-            }
-        default:
-            log.warn("May not be able collect summary statistics in file " + INPUT.getAbsoluteFile() +
-            " because it is not sorted in coordinate order.  If any of the reads are aligned this will blow up.");
-        }
     }
 
     /** Converts the supplied adapter sequences to byte arrays in both fwd and rc. */
@@ -207,7 +172,7 @@ public class CollectAlignmentSummaryMetrics extends CommandLineProgram {
         ADAPTER_SEQUENCES = new byte[count * 2][];
 
         for (int i=0; i<count; ++i) {
-            String adapter = ADAPTER_SEQUENCE.get(i).toUpperCase();
+            final String adapter = ADAPTER_SEQUENCE.get(i).toUpperCase();
             ADAPTER_SEQUENCES[i] = StringUtil.stringToBytes(adapter);
             ADAPTER_SEQUENCES[i + count] = StringUtil.stringToBytes(SequenceUtil.reverseComplement(adapter));
 
@@ -222,6 +187,7 @@ public class CollectAlignmentSummaryMetrics extends CommandLineProgram {
      * @return true if the read matches an adapter and false otherwise 
      */
     protected boolean isAdapterSequence(final byte[] read) {
+        if (read.length < ADAPTER_MATCH_LENGTH) return false;
         StringUtil.toUpperCase(read);
         
         for (final byte[] adapter : ADAPTER_SEQUENCES) {
@@ -241,31 +207,69 @@ public class CollectAlignmentSummaryMetrics extends CommandLineProgram {
         return false;
     }
 
-    /** Constructs a metrics collector for the supplied category. */
-    private MetricCollector<AlignmentSummaryMetrics, SAMRecord> constructCollector(final Category category) {
-        final MetricCollector<AlignmentSummaryMetrics, SAMRecord> collector =
-            new AggregateMetricCollector<AlignmentSummaryMetrics, SAMRecord>(new ReadCounter(), new QualityMappingCounter());
-        collector.setMetrics(new AlignmentSummaryMetrics());
-        collector.getMetrics().CATEGORY = category;
-        return collector;
-    }
-
     /**
      * Class that counts reads that match various conditions
      */
-    private class ReadCounter implements MetricCollector<AlignmentSummaryMetrics, SAMRecord> {
+    private class AlignmentSummaryMetricsCollector {
         private long numPositiveStrand = 0;
         private final Histogram<Integer> readLengthHistogram = new Histogram<Integer>();
         private AlignmentSummaryMetrics metrics;
         private long chimeras;
         private long adapterReads;
 
-        public void addRecord(final SAMRecord record) {
+        private long nonBisulfiteAlignedBases = 0;
+        private long hqNonBisulfiteAlignedBases = 0;
+        private final Histogram<Long> mismatchHistogram = new Histogram<Long>();
+        private final Histogram<Long> hqMismatchHistogram = new Histogram<Long>();
+        private final Histogram<Integer> badCycleHistogram = new Histogram<Integer>();
+
+        public AlignmentSummaryMetricsCollector(Category pairingCategory) {
+            metrics = new AlignmentSummaryMetrics();
+            metrics.CATEGORY = pairingCategory;
+        }
+
+        public void addRecord(final SAMRecord record, final ReferenceSequence ref) {
             if (record.getNotPrimaryAlignmentFlag()) {
                 // only want 1 count per read so skip non primary alignments
                 return;
             }
 
+            collectReadData(record, ref);
+            collectQualityData(record, ref);
+        }
+
+        public void onComplete() {
+
+            //summarize read data
+            if (metrics.TOTAL_READS > 0)
+            {
+                metrics.PCT_PF_READS = (double) metrics.PF_READS / (double) metrics.TOTAL_READS;
+                metrics.PCT_ADAPTER = this.adapterReads / (double) metrics.PF_READS;
+                metrics.MEAN_READ_LENGTH = readLengthHistogram.getMean();
+
+                //Calculate BAD_CYCLES
+                metrics.BAD_CYCLES = 0;
+                for (final Histogram<Integer>.Bin cycleBin : badCycleHistogram.values()) {
+                    final double badCyclePercentage = cycleBin.getValue() / metrics.TOTAL_READS;
+                    if (badCyclePercentage >= .8) {
+                        metrics.BAD_CYCLES++;
+                    }
+                }
+
+                if(doRefMetrics) {
+                    metrics.PCT_PF_READS_ALIGNED = (double) metrics.PF_READS_ALIGNED / (double) metrics.PF_READS;
+                    metrics.PCT_READS_ALIGNED_IN_PAIRS = (double) metrics.READS_ALIGNED_IN_PAIRS/ (double) metrics.PF_READS_ALIGNED;
+                    metrics.STRAND_BALANCE = numPositiveStrand / (double) metrics.PF_READS_ALIGNED;
+                    metrics.PCT_CHIMERAS = this.chimeras / (double) metrics.PF_HQ_ALIGNED_READS;
+
+                    metrics.PF_MISMATCH_RATE = mismatchHistogram.getSum() / (double) nonBisulfiteAlignedBases;
+                    metrics.PF_HQ_MEDIAN_MISMATCHES = hqMismatchHistogram.getMedian();
+                    metrics.PF_HQ_ERROR_RATE = hqMismatchHistogram.getSum() / (double) hqNonBisulfiteAlignedBases;
+                }
+            }
+        }
+
+        private void collectReadData(final SAMRecord record, final ReferenceSequence ref) {
             metrics.TOTAL_READS++;
             readLengthHistogram.increment(record.getReadBases().length);
 
@@ -282,24 +286,26 @@ public class CollectAlignmentSummaryMetrics extends CommandLineProgram {
                     }
                 }
                 else {
-                    metrics.PF_READS_ALIGNED++;
+                    if(doRefMetrics) {
+                        metrics.PF_READS_ALIGNED++;
 
-                    if (!record.getReadNegativeStrandFlag()) {
-                        numPositiveStrand++;
-                    }
+                        if (!record.getReadNegativeStrandFlag()) {
+                            numPositiveStrand++;
+                        }
 
-                    if (record.getReadPairedFlag() && !record.getMateUnmappedFlag()) {
-                        metrics.READS_ALIGNED_IN_PAIRS++;
+                        if (record.getReadPairedFlag() && !record.getMateUnmappedFlag()) {
+                            metrics.READS_ALIGNED_IN_PAIRS++;
 
-                        // With both reads mapped we can see if this pair is chimeric
-                        if (Math.abs(record.getInferredInsertSize()) > MAX_INSERT_SIZE ||
-                             !record.getReferenceIndex().equals(record.getMateReferenceIndex())) {
+                            // With both reads mapped we can see if this pair is chimeric
+                            if (Math.abs(record.getInferredInsertSize()) > MAX_INSERT_SIZE ||
+                                 !record.getReferenceIndex().equals(record.getMateReferenceIndex())) {
 
-                            // Check that both ends have mapq > minimum
-                            final Integer mateMq = record.getIntegerAttribute("MQ");
-                            if (mateMq == null || mateMq >= MAPPING_QUALITY_THRESHOLD &&
-                                    record.getMappingQuality() >= MAPPING_QUALITY_THRESHOLD) {
-                                ++this.chimeras;
+                                // Check that both ends have mapq > minimum
+                                final Integer mateMq = record.getIntegerAttribute("MQ");
+                                if (mateMq == null || mateMq >= MAPPING_QUALITY_THRESHOLD &&
+                                        record.getMappingQuality() >= MAPPING_QUALITY_THRESHOLD) {
+                                    ++this.chimeras;
+                                }
                             }
                         }
                     }
@@ -307,43 +313,9 @@ public class CollectAlignmentSummaryMetrics extends CommandLineProgram {
             }
         }
 
-        public void onComplete() {
-            metrics.PCT_PF_READS = (double) metrics.PF_READS / (double) metrics.TOTAL_READS;
-            metrics.PCT_PF_READS_ALIGNED = (double) metrics.PF_READS_ALIGNED / (double) metrics.PF_READS;
-            metrics.PCT_READS_ALIGNED_IN_PAIRS = (double) metrics.READS_ALIGNED_IN_PAIRS/ (double) metrics.PF_READS_ALIGNED;
-            metrics.MEAN_READ_LENGTH = readLengthHistogram.getMean();
-            metrics.STRAND_BALANCE = numPositiveStrand / (double) metrics.PF_READS_ALIGNED;
-            metrics.PCT_ADAPTER = this.adapterReads / (double) metrics.PF_READS;
-            metrics.PCT_CHIMERAS = this.chimeras / (double) metrics.PF_HQ_ALIGNED_READS;
-        }
-
-        private boolean isNoiseRead(final SAMRecord record) {
-            final Object noiseAttribute = record.getAttribute(ReservedTagConstants.XN);
-            return (noiseAttribute != null && noiseAttribute.equals(1));
-        }
-
-        public void setMetrics(final AlignmentSummaryMetrics metrics) {
-            this.metrics = metrics;
-        }
-
-        public AlignmentSummaryMetrics getMetrics() {
-            return this.metrics;
-        }
-    }
-
-    /**
-     * Class that counts quality mappings & base calls that match various conditions
-     */
-    private class QualityMappingCounter implements MetricCollector<AlignmentSummaryMetrics, SAMRecord> {
-        private final Histogram<Long> mismatchHistogram = new Histogram<Long>();
-        private final Histogram<Integer> badCycleHistogram = new Histogram<Integer>();
-        private AlignmentSummaryMetrics metrics;
-
-        public void addRecord(final SAMRecord record) {
-            if (record.getNotPrimaryAlignmentFlag()) {
-                return;
-            }
-            if (record.getReadUnmappedFlag()) {
+        private void collectQualityData(final SAMRecord record, final ReferenceSequence reference)
+        {
+            if (record.getReadUnmappedFlag() || !doRefMetrics) {
                 final byte[] readBases = record.getReadBases();
                 for (int i = 0; i < readBases.length; i++) {
                     if (SequenceUtil.isNoCall(readBases[i])) {
@@ -353,13 +325,14 @@ public class CollectAlignmentSummaryMetrics extends CommandLineProgram {
             } else {
                 final boolean highQualityMapping = isHighQualityMapping(record);
                 if (highQualityMapping) metrics.PF_HQ_ALIGNED_READS++;
-                
+
                 final byte[] readBases = record.getReadBases();
-                final byte[] refBases = referenceSequenceWalker.get(record.getReferenceIndex()).getBases();
+                final byte[] refBases = reference.getBases();
                 final byte[] qualities  = record.getBaseQualities();
                 final int refLength = refBases.length;
-                long mismatchCount = 0;
-                
+                long mismatchCount   = 0;
+                long hqMismatchCount = 0;
+
                 for (final AlignmentBlock alignmentBlock : record.getAlignmentBlocks()) {
                     final int readIndex = alignmentBlock.getReadStart() - 1;
                     final int refIndex  = alignmentBlock.getReferenceStart() - 1;
@@ -381,16 +354,24 @@ public class CollectAlignmentSummaryMetrics extends CommandLineProgram {
                                 mismatch = false;
                             }
                         }
+
+                        if(mismatch) mismatchCount++;
+
+                        metrics.PF_ALIGNED_BASES++;
+                        if(!bisulfiteBase) {
+                            nonBisulfiteAlignedBases++;
+                        }
+                        
                         if (highQualityMapping) {
                             metrics.PF_HQ_ALIGNED_BASES++;
                             if (!bisulfiteBase) {
-                                metrics.incrementErrorRateDenominator();                                
+                                hqNonBisulfiteAlignedBases++;
                             }
                             if (qualities[readBaseIndex] >= BASE_QUALITY_THRESHOLD) {
                                 metrics.PF_HQ_ALIGNED_Q20_BASES++;
                             }
                             if (mismatch) {
-                                mismatchCount++;
+                                hqMismatchCount++;
                             }
                         }
                         if (mismatch || SequenceUtil.isNoCall(readBases[readBaseIndex])) {
@@ -400,31 +381,20 @@ public class CollectAlignmentSummaryMetrics extends CommandLineProgram {
                 }
 
                 mismatchHistogram.increment(mismatchCount);
+                hqMismatchHistogram.increment(hqMismatchCount);
             }
         }
 
+        private boolean isNoiseRead(final SAMRecord record) {
+            final Object noiseAttribute = record.getAttribute(ReservedTagConstants.XN);
+            return (noiseAttribute != null && noiseAttribute.equals(1));
+        }
+
         private boolean isHighQualityMapping(final SAMRecord record) {
             return !record.getReadFailsVendorQualityCheckFlag() &&
             record.getMappingQuality() >= MAPPING_QUALITY_THRESHOLD;
         }
 
-        public void onComplete() {
-            metrics.PF_HQ_MEDIAN_MISMATCHES = mismatchHistogram.getMedian();
-            metrics.PF_HQ_ERROR_RATE = mismatchHistogram.getSum() / (double)metrics.getErrorRateDenominator();
-            metrics.BAD_CYCLES = 0;
-
-            for (final Histogram<Integer>.Bin cycleBin : badCycleHistogram.values()) {
-                final double badCyclePercentage = cycleBin.getValue() / metrics.TOTAL_READS;
-                if (badCyclePercentage >= .8) {
-                    metrics.BAD_CYCLES++;
-                }
-            }
-        }
-
-        public void setMetrics(final AlignmentSummaryMetrics metrics) {
-            this.metrics = metrics;
-        }
-
         public AlignmentSummaryMetrics getMetrics() {
             return this.metrics;
         }
