@@ -83,7 +83,7 @@ public class CommandLineParser {
      * public String USAGE = CommandLineParser.getStandardUsagePreamble(getClass()) + "Frobnicates the freebozzle."
      */
     public static String getStandardUsagePreamble(final Class mainClass) {
-        return "USAGE: " + mainClass.getName() + " [options]\n\n";
+        return "USAGE: " + mainClass.getSimpleName() + " [options]\n\n";
     }
 
     // This is the object that the caller has provided that contains annotations,
@@ -128,7 +128,7 @@ public class CommandLineParser {
     public CommandLineParser(final Object callerOptions) {
         this.callerOptions = callerOptions;
 
-        for (final Field field : this.callerOptions.getClass().getFields()) {
+        for (final Field field : getAllFields(this.callerOptions.getClass())) {
             if (field.getAnnotation(PositionalArguments.class) != null) {
                 handlePositionalArgumentAnnotation(field);
             }
@@ -149,58 +149,79 @@ public class CommandLineParser {
         }
     }
 
+    private List<Field> getAllFields(Class clazz) {
+        final List<Field> ret = new ArrayList<Field>();
+        do {
+            ret.addAll(Arrays.asList(clazz.getDeclaredFields()));
+            clazz = clazz.getSuperclass();
+        } while (clazz != null);
+        return ret;
+    }
+
     /**
      * Print a usage message based on the options object passed to the ctor.
      * @param stream Where to write the usage message.
      */
-    public void usage(final PrintStream stream) {
+    public void usage(final PrintStream stream, final boolean printCommon) {
         stream.print(usagePreamble);
+        stream.println("\n\nOptions:\n");
+
+        printOptionParamUsage(stream, "--help", "-h", null, "Displays options specific to this tool.");
+        printOptionParamUsage(stream, "--stdhelp", "-H", null, "Displays options specific to this tool AND " +
+                "options common to all Picard command line tools.");
+
         if (!optionDefinitions.isEmpty()) {
-            stream.println("\n\nOptions:\n");
             for (final OptionDefinition optionDefinition : optionDefinitions) {
-                printOptionUsage(stream, optionDefinition);
+                if(printCommon || !optionDefinition.isCommon) printOptionUsage(stream, optionDefinition);
             }
         }
-        final Field fileField;
-        try {
-            fileField = getClass().getField("IGNORE_THIS_PROPERTY");
-        } catch (NoSuchFieldException e) {
-            throw new PicardException("Should never happen", e);
+
+        if(printCommon) {
+            final Field fileField;
+            try {
+                fileField = getClass().getField("IGNORE_THIS_PROPERTY");
+            } catch (NoSuchFieldException e) {
+                throw new PicardException("Should never happen", e);
+            }
+            final OptionDefinition optionsFileOptionDefinition =
+                    new OptionDefinition(fileField, OPTIONS_FILE, "",
+                            "File of OPTION_NAME=value pairs.  No positional parameters allowed.  Unlike command-line options, " +
+                    "unrecognized options are ignored.  " + "A single-valued option set in an options file may be overridden " +
+                    "by a subsequent command-line option.  " +
+                    "A line starting with '#' is considered a comment.", false, true, 0, Integer.MAX_VALUE, null, true, new String[0]);
+            printOptionUsage(stream, optionsFileOptionDefinition);
         }
-        final OptionDefinition optionsFileOptionDefinition =
-                new OptionDefinition(fileField, OPTIONS_FILE, "",
-                        "File of OPTION_NAME=value pairs.  No positional parameters allowed.  Unlike command-line options, " +
-                "unrecognized options are ignored.  " + "A single-valued option set in an options file may be overridden " +
-                "by a subsequent command-line option.  " +
-                "A line starting with '#' is considered a comment.", false, true, 0, Integer.MAX_VALUE, null, new String[0]);
-        printOptionUsage(stream, optionsFileOptionDefinition);
     }
 
-    public void htmlUsage(final PrintStream stream, final String programName) {
+    public void htmlUsage(final PrintStream stream, final String programName, boolean printCommon) {
         // TODO: Should HTML escape usage preamble and option usage, including line breaks
         stream.println("<a name=\"" + programName + "\"/>");
         stream.println("<h3>" + programName + "</h3>");
         stream.println("<p>" + htmlEscape(usagePreamble) + "</p>");
         boolean hasOptions = false;
         for (final OptionDefinition optionDefinition : optionDefinitions) {
-            if (!isStandardOption(optionDefinition)) {
+            if (!optionDefinition.isCommon || printCommon) {
                 hasOptions = true;
                 break;
             }
         }
         if (hasOptions) {
-            stream.println("<table>");
-            stream.println("<tr><th>Option</th><th>Description</th></tr>");
-            for (final OptionDefinition optionDefinition : optionDefinitions) {
-                if (!isStandardOption(optionDefinition)) {
-                    printHtmlOptionUsage(stream, optionDefinition);
-                }
-            }
-            stream.println("</table>");
+            htmlPrintOptions(stream, printCommon);
         }
         stream.println("<br/>");
     }
 
+    public void htmlPrintOptions(PrintStream stream, boolean printCommon) {
+        stream.println("<table>");
+        stream.println("<tr><th>Option</th><th>Description</th></tr>");
+        for (final OptionDefinition optionDefinition : optionDefinitions) {
+            if (!optionDefinition.isCommon || printCommon) {
+                printHtmlOptionUsage(stream, optionDefinition);
+            }
+        }
+        stream.println("</table>");
+    }
+
     private static String htmlEscape(String str) {
         // May need more here
         str = str.replaceAll("<", "&lt;");
@@ -208,15 +229,6 @@ public class CommandLineParser {
         return str;
     }
 
-    /**
-     * Override this method to suppress generating HTML help for "standard" options.
-     * @param optionDefinition
-     * @return true if HTML help should not be generated.
-     */
-    protected boolean isStandardOption(final OptionDefinition optionDefinition) {
-        return false;
-    }
-
     /**
      * Parse command-line options, and store values in callerOptions object passed to ctor.
      * @param messageStream Where to write error messages.
@@ -230,9 +242,14 @@ public class CommandLineParser {
         for (int i = 0; i < args.length; ++i) {
             final String arg = args[i];
             if (arg.equals("-h") || arg.equals("--help")) {
-                usage(messageStream);
+                usage(messageStream, false);
                 return false;
             }
+            if (arg.equals("-H") || arg.equals("--stdhelp")) {
+                usage(messageStream, true);
+                return false;
+            }
+
             final String[] pair = arg.split("=", 2);
             if (pair.length == 2 && pair[1].length() == 0) {
 
@@ -244,26 +261,26 @@ public class CommandLineParser {
                 if (pair[0].equals(OPTIONS_FILE)) {
                     if (!parseOptionsFile(pair[1])) {
                         messageStream.println();
-                        usage(messageStream);
+                        usage(messageStream, true);
                         return false;
                     }
                     commandLine += " " + OPTIONS_FILE + "=" + pair[1];
                 } else {
                     if (!parseOption(pair[0], pair[1], false)) {
                         messageStream.println();
-                        usage(messageStream);
+                        usage(messageStream, true);
                         return false;
                     }
                 }
             } else if (!parsePositionalArgument(arg)) {
                 messageStream.println();
-                usage(messageStream);
+                usage(messageStream, false);
                 return false;
             }
         }
         if (!checkNumArguments()) {
             messageStream.println();
-            usage(messageStream);
+            usage(messageStream, false);
             return false;
         }
 
@@ -405,7 +422,7 @@ public class CommandLineParser {
                 messageStream.println("ERROR: Option '" + key + "' cannot be specified more than once.");
                 return false;
             }
-        }
+        }                                      
         final Object value;
         try {
             if(stringValue.equals("null")) {
@@ -439,6 +456,8 @@ public class CommandLineParser {
                 } else {
                     c.add(value);
                 }
+                optionDefinition.hasBeenSet = true;
+                optionDefinition.hasBeenSetFromOptionsFile = optionsFile;
             } else {
                 optionDefinition.field.set(callerOptions, value);
                 optionDefinition.hasBeenSet = true;
@@ -465,19 +484,19 @@ public class CommandLineParser {
             reader = new BufferedReader(new FileReader(optionsFile));
             String line;
             while ((line = reader.readLine()) != null) {
-                if (line.startsWith("#")) {
+                if (line.startsWith("#") || line.trim().length() == 0) {
                     continue;
                 }
                 final String[] pair = line.split("=", 2);
                 if (pair.length == 2) {
                     if (!parseOption(pair[0], pair[1], true)) {
                         messageStream.println();
-                        usage(messageStream);
+                        usage(messageStream, true);
                         return false;
                     }
                 } else {
                     messageStream.println("Strange line in OPTIONS_FILE " + optionsFile + ": " + line);
-                    usage(messageStream);
+                    usage(messageStream, true);
                     return false;
                 }
             }
@@ -499,23 +518,31 @@ public class CommandLineParser {
     }
 
     private void printOptionUsage(final PrintStream stream, final OptionDefinition optionDefinition) {
-        final String type = getUnderlyingType(optionDefinition.field).getSimpleName();
-        String optionLabel = optionDefinition.name + "=" + type;
+        printOptionParamUsage(stream, optionDefinition.name, optionDefinition.shortName,
+                              getUnderlyingType(optionDefinition.field).getSimpleName(),
+                              makeOptionDescription(optionDefinition));
+    }
+
+
+    private void printOptionParamUsage(final PrintStream stream, final String name, final String shortName,
+                                       final String type, final String optionDescription) {
+        String optionLabel = name;
+        if(type != null) optionLabel += "=" + type;
+
         stream.print(optionLabel);
-        if (optionDefinition.shortName.length() > 0) {
+        if (shortName != null && shortName.length() > 0) {
             stream.println();
-        }
-        if (optionDefinition.shortName.length() > 0) {
-            optionLabel = optionDefinition.shortName + "=" + type;
+            optionLabel = shortName;
+            if(type != null) optionLabel += "=" + type;
             stream.print(optionLabel);
         }
+
         int numSpaces = OPTION_COLUMN_WIDTH - optionLabel.length();
         if (optionLabel.length() > OPTION_COLUMN_WIDTH) {
             stream.println();
             numSpaces = OPTION_COLUMN_WIDTH;
         }
         printSpaces(stream, numSpaces);
-        final String optionDescription = makeOptionDescription(optionDefinition);
         final String wrappedDescription = StringUtil.wordWrap(optionDescription, DESCRIPTION_COLUMN_WIDTH);
         final String[] descriptionLines = wrappedDescription.split("\n");
         for (int i = 0; i < descriptionLines.length; ++i) {
@@ -599,6 +626,7 @@ public class CommandLineParser {
 
     private void handleOptionAnnotation(final Field field) {
         try {
+            field.setAccessible(true);
             final Option optionAnnotation = field.getAnnotation(Option.class);
             final boolean isCollection = isCollectionField(field);
             if (isCollection) {
@@ -624,7 +652,7 @@ public class CommandLineParser {
                     optionAnnotation.shortName(),
                     optionAnnotation.doc(), optionAnnotation.optional() || (field.get(callerOptions) != null),
                     isCollection, optionAnnotation.minElements(),
-                    optionAnnotation.maxElements(), field.get(callerOptions),
+                    optionAnnotation.maxElements(), field.get(callerOptions),  optionAnnotation.common(),
                     optionAnnotation.mutex());
 
             for (final String option : optionAnnotation.mutex()) {
@@ -656,6 +684,7 @@ public class CommandLineParser {
                     ("@Usage cannot be used more than once in an option class.");
         }
         try {
+            field.setAccessible(true);
             usagePreamble = (String)field.get(callerOptions);
             final Usage usageAnnotation = field.getAnnotation(Usage.class);
             if (usageAnnotation.programVersion().length() > 0) {
@@ -809,12 +838,13 @@ public class CommandLineParser {
         final int minElements;
         final int maxElements;
         final String defaultValue;
+        final boolean isCommon;
         boolean hasBeenSet = false;
         boolean hasBeenSetFromOptionsFile = false;
         Set<String> mutuallyExclusive;
 
         private OptionDefinition(final Field field, final String name, final String shortName, final String doc, final boolean optional, final boolean collection,
-                                 final int minElements, final int maxElements, final Object defaultValue, final String[] mutuallyExclusive) {
+                                 final int minElements, final int maxElements, final Object defaultValue, final boolean isCommon, final String[] mutuallyExclusive) {
             this.field = field;
             this.name = name.toUpperCase();
             this.shortName = shortName.toUpperCase();
@@ -834,6 +864,7 @@ public class CommandLineParser {
             } else {
                 this.defaultValue = "null";
             }
+            this.isCommon = isCommon;
             this.mutuallyExclusive = new HashSet<String>(Arrays.asList(mutuallyExclusive));
         }
     }
