@@ -31,10 +31,7 @@ import net.sf.samtools.util.StringUtil;
 
 import java.io.*;
 import java.lang.reflect.Field;
-import java.util.ArrayList;
-import java.util.Collections;
-import java.util.List;
-import java.util.TreeSet;
+import java.util.*;
 
 /**
  * Contains a set of metrics that can be written to a file and parsed back
@@ -218,23 +215,28 @@ public class MetricsFile<BEAN extends MetricBase, HKEY extends Comparable> {
 
     /** Prints the histogram if one is present. */
     private void printHistogram(BufferedWriter out, FormatUtil formatter) throws IOException {
-        if (this.histograms.isEmpty()) {
+        final List<Histogram<HKEY>> nonEmptyHistograms = new ArrayList<Histogram<HKEY>>();
+        for (Histogram<HKEY> histo : this.histograms) {
+            if (!histo.isEmpty()) nonEmptyHistograms.add(histo);
+        }
+
+        if (nonEmptyHistograms.isEmpty()) {
             return;
         }
 
         // Build a combined key set.  Assume comparator is the same for all Histograms
-        java.util.Set<HKEY> keys = new TreeSet<HKEY>(histograms.get(0).comparator());
-        for (Histogram<HKEY> histo : histograms) {
+        java.util.Set<HKEY> keys = new TreeSet<HKEY>(nonEmptyHistograms.get(0).comparator());
+        for (Histogram<HKEY> histo : nonEmptyHistograms) {
             if (histo != null) keys.addAll(histo.keySet());
         }
 
         // Add a header for the histogram key type
-        out.append(HISTO_HEADER + this.histograms.get(0).keySet().iterator().next().getClass().getName());
+        out.append(HISTO_HEADER + nonEmptyHistograms.get(0).keySet().iterator().next().getClass().getName());
         out.newLine();
 
         // Output a header row
-        out.append(StringUtil.assertCharactersNotInString(this.histograms.get(0).getBinLabel(), '\t', '\n'));
-        for (Histogram<HKEY> histo : this.histograms) {
+        out.append(StringUtil.assertCharactersNotInString(nonEmptyHistograms.get(0).getBinLabel(), '\t', '\n'));
+        for (Histogram<HKEY> histo : nonEmptyHistograms) {
             out.append(SEPARATOR);
             out.append(StringUtil.assertCharactersNotInString(histo.getValueLabel(), '\t', '\n'));
         }
@@ -243,7 +245,7 @@ public class MetricsFile<BEAN extends MetricBase, HKEY extends Comparable> {
         for (HKEY key : keys) {
             out.append(key.toString());
 
-            for (Histogram<HKEY> histo : this.histograms) {
+            for (Histogram<HKEY> histo : nonEmptyHistograms) {
                 Histogram<HKEY>.Bin bin = histo.get(key);
                 final double value = (bin == null ? 0 : bin.getValue());
 
@@ -288,7 +290,7 @@ public class MetricsFile<BEAN extends MetricBase, HKEY extends Comparable> {
                     
                     String className = line.substring(MAJOR_HEADER_PREFIX.length()).trim();
                     try {
-                        header = (Header) loadClass(className).newInstance();
+                        header = (Header) loadClass(className, true).newInstance();
                     }
                     catch (Exception e) {
                         throw new PicardException("Error load and/or instantiating an instance of " + className, e);
@@ -307,63 +309,67 @@ public class MetricsFile<BEAN extends MetricBase, HKEY extends Comparable> {
                 }
             }
 
-            if (line == null) {
-                throw new PicardException("No lines in metrics file after header.");
+            //read space between starting headers and metrics
+            while (line != null && !line.startsWith(MAJOR_HEADER_PREFIX)) {
+                line = in.readLine();
             }
-            // Then read the metrics if there are any
-            while (!line.startsWith(MAJOR_HEADER_PREFIX)) {
-                line = in.readLine().trim();
-            }
-            if (line.startsWith(METRIC_HEADER)) {
-                // Get the metric class from the header
-                String className = line.split(SEPARATOR)[1];
-                Class<?> type = null;
-                try {
-                    type = loadClass(className);
-                }
-                catch (ClassNotFoundException cnfe) {
-                    throw new PicardException("Could not locate class with name " + className, cnfe);
-                }
 
-                // Read the next line with the column headers
-                String[] fieldNames = in.readLine().split(SEPARATOR);
-                Field[] fields = new Field[fieldNames.length];
-                for (int i=0; i<fieldNames.length; ++i) {
+
+            if(line != null) {
+                line = line.trim();
+            
+                // Then read the metrics if there are any
+                if (line.startsWith(METRIC_HEADER)) {
+                    // Get the metric class from the header
+                    String className = line.split(SEPARATOR)[1];
+                    Class<?> type = null;
                     try {
-                        fields[i] = type.getField(fieldNames[i]);
+                        type = loadClass(className, true);
                     }
-                    catch (Exception e) {
-                        throw new PicardException("Could not get field with name " + fieldNames[i] +
-                            " from class " + type.getName());
+                    catch (ClassNotFoundException cnfe) {
+                        throw new PicardException("Could not locate class with name " + className, cnfe);
                     }
-                }
 
-                // Now read the values
-                while ((line = in.readLine()) != null) {
-                    if ("".equals(line.trim())) {
-                        break;
-                    }
-                    else {
-                        String[] values = line.split(SEPARATOR, -1);
-                        BEAN bean = null;
-
-                        try { bean = (BEAN) type.newInstance(); }
-                        catch (Exception e) { throw new PicardException("Error instantiating a " + type.getName(), e); }
-
-                        for (int i=0; i<fields.length; ++i) {
-                            Object value = null;
-                            if (values[i] != null && values[i].length() > 0) {
-                                value = formatter.parseObject(values[i], fields[i].getType());
-                            }
-
-                            try { fields[i].set(bean, value); }
-                            catch (Exception e) {
-                                throw new PicardException("Error setting field " + fields[i].getName() +
-                                        " on class of type " + type.getName(), e);
-                            }
+                    // Read the next line with the column headers
+                    String[] fieldNames = in.readLine().split(SEPARATOR);
+                    Field[] fields = new Field[fieldNames.length];
+                    for (int i=0; i<fieldNames.length; ++i) {
+                        try {
+                            fields[i] = type.getField(fieldNames[i]);
                         }
+                        catch (Exception e) {
+                            throw new PicardException("Could not get field with name " + fieldNames[i] +
+                                " from class " + type.getName());
+                        }
+                    }
 
-                        this.metrics.add(bean);
+                    // Now read the values
+                    while ((line = in.readLine()) != null) {
+                        if ("".equals(line.trim())) {
+                            break;
+                        }
+                        else {
+                            String[] values = line.split(SEPARATOR, -1);
+                            BEAN bean = null;
+
+                            try { bean = (BEAN) type.newInstance(); }
+                            catch (Exception e) { throw new PicardException("Error instantiating a " + type.getName(), e); }
+
+                            for (int i=0; i<fields.length; ++i) {
+                                Object value = null;
+                                if (values[i] != null && values[i].length() > 0) {
+                                    value = formatter.parseObject(values[i], fields[i].getType());
+                                }
+
+                                try { fields[i].set(bean, value); }
+                                catch (Exception e) {
+                                    throw new PicardException("Error setting field " + fields[i].getName() +
+                                            " on class of type " + type.getName(), e);
+                                }
+                            }
+
+                            this.metrics.add(bean);
+                        }
                     }
                 }
             }
@@ -377,7 +383,7 @@ public class MetricsFile<BEAN extends MetricBase, HKEY extends Comparable> {
                 String keyClassName = line.split(SEPARATOR)[1].trim();
                 Class<?> keyClass = null;
 
-                try { keyClass = loadClass(keyClassName); }
+                try { keyClass = loadClass(keyClassName, true); }
                 catch (ClassNotFoundException cnfe) { throw new PicardException("Could not load class with name " + keyClassName); }
 
                 // Read the next line with the bin and value labels
@@ -404,17 +410,41 @@ public class MetricsFile<BEAN extends MetricBase, HKEY extends Comparable> {
     }
 
     /** Attempts to load a class, taking into account that some classes have "migrated" from the broad to sf. */
-    private Class<?> loadClass(String className) throws ClassNotFoundException {
-        try {
-            return Class.forName(className);
-        }
+    private Class<?> loadClass(String className, boolean tryOtherPackages) throws ClassNotFoundException {
+        // List of alternative packages to check in case classes moved around
+        String[] packages = new String[] {
+                "edu.mit.broad.picard.genotype.concordance",
+                "edu.mit.broad.picard.genotype.fingerprint",
+                "edu.mit.broad.picard.ic",
+                "edu.mit.broad.picard.illumina",
+                "edu.mit.broad.picard.jumping",
+                "edu.mit.broad.picard.quality",
+                "edu.mit.broad.picard.samplevalidation",
+                "net.sf.picard.analysis",
+                "net.sf.picard.analysis.directed",
+                "net.sf.picard.sam",
+                "net.sf.picard.metrics"
+        };
+
+        try { return Class.forName(className); }
         catch (ClassNotFoundException cnfe) {
-            if (className.startsWith("edu.mit.broad.picard")) {
-                return loadClass(className.replace("edu.mit.broad.picard", "net.sf.picard"));
-            }
-            else {
-                throw cnfe;
+            if (tryOtherPackages) {
+                for (final String p : packages) {
+                    try {
+                        return loadClass(p + className.substring(className.lastIndexOf(".")), false);
+                    }
+                    catch (ClassNotFoundException cnf2) {/* do nothing */}
+                    // If it ws an inner class, try and see if it's a stand-alone class now
+                    if (className.indexOf("$") > -1) {
+                        try {
+                            return loadClass(p + "." + className.substring(className.lastIndexOf("$") + 1), false);
+                        }
+                        catch (ClassNotFoundException cnf2) {/* do nothing */}
+                    }
+                }
             }
+
+            throw cnfe;
         }
     }
 
@@ -429,19 +459,31 @@ public class MetricsFile<BEAN extends MetricBase, HKEY extends Comparable> {
         }
         MetricsFile that = (MetricsFile) o;
 
-        if (!this.headers.equals(that.headers)) {
+        if (!areHeadersEqual(that)) {
             return false;
         }
-        if (!this.metrics.equals(that.metrics)) {
+        if (!areMetricsEqual(that)) {
             return false;
         }
-        if (!this.histograms.equals(that.histograms)) {
+        if (!areHistogramsEqual(that)) {
             return false;
         }
 
         return true;
     }
 
+    public boolean areHeadersEqual(MetricsFile that) {
+        return this.headers.equals(that.headers);
+    }
+
+    public boolean areMetricsEqual(MetricsFile that) {
+        return this.metrics.equals(that.metrics);
+    }
+
+    public boolean areHistogramsEqual(MetricsFile that) {
+        return this.histograms.equals(that.histograms);
+    }
+
     @Override
     public int hashCode() {
         int result = headers.hashCode();
