@@ -73,10 +73,14 @@ public class MarkDuplicates extends AbstractDuplicateFindingAlgorithm {
             shortName=StandardOptionDefinitions.ASSUME_SORTED_SHORT_NAME)
     public boolean ASSUME_SORTED = false;
 
-    @Option(doc="The maximum number of sequences allowed in SAM file.  If this value is exceeded, program won't " +
-            "spill to disk (used to avoid situation where there are not enough file handles.", shortName="MAX_SEQS")
+    @Option(doc="This option is obsolete.  ReadEnds will always be spilled to disk.", shortName="MAX_SEQS")
     public int MAX_SEQUENCES_FOR_DISK_READ_ENDS_MAP = 50000;
 
+    @Option(doc="Maximum number of file handles to keep open when spilling read ends to disk.  " + "" +
+            "Set this number a little lower than the per-process maximum number of file that may be open.  " +
+            "This number can be found by executing the 'ulimit -n' command on a Unix system.", shortName = "MAX_FILE_HANDLES")
+    public int MAX_FILE_HANDLES_FOR_READ_ENDS_MAP = 8000;
+
     private SortingCollection<ReadEnds> pairSort;
     private SortingCollection<ReadEnds> fragSort;
     private SortingLongCollection duplicateIndexes;
@@ -126,49 +130,64 @@ public class MarkDuplicates extends AbstractDuplicateFindingAlgorithm {
         long recordInFileIndex = 0;
         long nextDuplicateIndex = (this.duplicateIndexes.hasNext() ? this.duplicateIndexes.next(): -1);
 
+        if(header != null) {
+            for(final SAMReadGroupRecord readGroup : header.getReadGroups()) {
+                final String library = readGroup.getLibrary();
+                DuplicationMetrics metrics = metricsByLibrary.get(library);
+                if (metrics == null) {
+                    metrics = new DuplicationMetrics();
+                    metrics.LIBRARY = library;
+                    metricsByLibrary.put(library, metrics);
+                }
+            }
+        }
+
         for (final SAMRecord rec : in) {
-            final String library = getLibraryName(header, rec);
-            DuplicationMetrics metrics = metricsByLibrary.get(library);
-            if (metrics == null) {
-                metrics = new DuplicationMetrics();
-                metrics.LIBRARY = library;
-                metricsByLibrary.put(library, metrics);
-            }
+            if (!rec.getNotPrimaryAlignmentFlag()) {
+                final String library = getLibraryName(header, rec);
+                DuplicationMetrics metrics = metricsByLibrary.get(library);
+                if (metrics == null) {
+                    metrics = new DuplicationMetrics();
+                    metrics.LIBRARY = library;
+                    metricsByLibrary.put(library, metrics);
+                }
 
-            // First bring the simple metrics up to date
-            if (rec.getReadUnmappedFlag()) {
-                ++metrics.UNMAPPED_READS;
-            }
-            else if (!rec.getReadPairedFlag() || rec.getMateUnmappedFlag()) {
-                ++metrics.UNPAIRED_READS_EXAMINED;
-            }
-            else {
-                ++metrics.READ_PAIRS_EXAMINED; // will need to be divided by 2 at the end
-            }
-
-
-            if (recordInFileIndex++ == nextDuplicateIndex) {
-                rec.setDuplicateReadFlag(true);
-
-                // Update the duplication metrics
-                if (!rec.getReadPairedFlag() || rec.getMateUnmappedFlag()) {
-                    ++metrics.UNPAIRED_READ_DUPLICATES;
+                // First bring the simple metrics up to date
+                if (rec.getReadUnmappedFlag()) {
+                    ++metrics.UNMAPPED_READS;
+                }
+                else if (!rec.getReadPairedFlag() || rec.getMateUnmappedFlag()) {
+                    ++metrics.UNPAIRED_READS_EXAMINED;
                 }
                 else {
-                    ++metrics.READ_PAIR_DUPLICATES;// will need to be divided by 2 at the end
+                    ++metrics.READ_PAIRS_EXAMINED; // will need to be divided by 2 at the end
                 }
 
-                // Now try and figure out the next duplicate index
-                if (this.duplicateIndexes.hasNext()) {
-                    nextDuplicateIndex = this.duplicateIndexes.next();
-                } else {
-                    // Only happens once we've marked all the duplicates
-                    nextDuplicateIndex = -1;
+
+                if (recordInFileIndex == nextDuplicateIndex) {
+                    rec.setDuplicateReadFlag(true);
+
+                    // Update the duplication metrics
+                    if (!rec.getReadPairedFlag() || rec.getMateUnmappedFlag()) {
+                        ++metrics.UNPAIRED_READ_DUPLICATES;
+                    }
+                    else {
+                        ++metrics.READ_PAIR_DUPLICATES;// will need to be divided by 2 at the end
+                    }
+
+                    // Now try and figure out the next duplicate index
+                    if (this.duplicateIndexes.hasNext()) {
+                        nextDuplicateIndex = this.duplicateIndexes.next();
+                    } else {
+                        // Only happens once we've marked all the duplicates
+                        nextDuplicateIndex = -1;
+                    }
+                }
+                else {
+                    rec.setDuplicateReadFlag(false);
                 }
             }
-            else {
-                rec.setDuplicateReadFlag(false);
-            }
+            recordInFileIndex++;
 
             if (this.REMOVE_DUPLICATES && rec.getDuplicateReadFlag()) {
                 // do nothing
@@ -178,6 +197,8 @@ public class MarkDuplicates extends AbstractDuplicateFindingAlgorithm {
             }
         }
 
+        this.duplicateIndexes.cleanup();
+
         reportMemoryStats("Before output close");
         out.close();
         reportMemoryStats("After output close");
@@ -248,12 +269,8 @@ public class MarkDuplicates extends AbstractDuplicateFindingAlgorithm {
                 throw new PicardException(INPUT + " is not coordinate sorted.");
             }
         }
-        final ReadEndsMap tmp;
-        if (header.getSequenceDictionary().getSequences().size() > MAX_SEQUENCES_FOR_DISK_READ_ENDS_MAP) {
-            tmp = new RAMReadEndsMap();
-        } else {
-            tmp = new DiskReadEndsMap();
-        }
+        final ReadEndsMap tmp = new DiskReadEndsMap(MAX_FILE_HANDLES_FOR_READ_ENDS_MAP);
+
         long index = 0;
 
         for (final SAMRecord rec : sam) {
@@ -264,7 +281,7 @@ public class MarkDuplicates extends AbstractDuplicateFindingAlgorithm {
                 }
                 // If this read is unmapped but sorted with the mapped reads, just skip it.
             }
-            else {
+            else if (!rec.getNotPrimaryAlignmentFlag()){
                 final ReadEnds fragmentEnd = buildReadEnds(header, index, rec);
                 this.fragSort.add(fragmentEnd);
 
@@ -449,6 +466,7 @@ public class MarkDuplicates extends AbstractDuplicateFindingAlgorithm {
             }
         }
         markDuplicatePairs(nextChunk);
+        this.pairSort.cleanup();
         this.pairSort = null;
 
         // Now deal with the fragments
@@ -475,6 +493,7 @@ public class MarkDuplicates extends AbstractDuplicateFindingAlgorithm {
             }
         }
         markDuplicateFragments(nextChunk, containsPairs);
+        this.fragSort.cleanup();
         this.fragSort = null;
 
         log.info("Sorting list of duplicate records.");
