@@ -31,6 +31,7 @@ import net.sf.picard.reference.ReferenceSequence;
 import net.sf.picard.reference.ReferenceSequenceFile;
 import net.sf.picard.reference.ReferenceSequenceFileWalker;
 import net.sf.picard.util.Histogram;
+import net.sf.picard.util.Log;
 import net.sf.samtools.*;
 import net.sf.samtools.SAMFileReader.ValidationStringency;
 import net.sf.samtools.SAMValidationError.Type;
@@ -59,7 +60,7 @@ import java.util.*;
 public class SamFileValidator {
     private Histogram<Type> errorsByType = new Histogram<Type>();
     private final PrintWriter out;
-    private Map<String, PairEndInfo> pairEndInfoByName;
+    private PairEndInfoMap pairEndInfoByName;
     private ReferenceSequenceFileWalker refFileWalker = null;
     private boolean verbose = false;
     private int maxVerboseOutput = 100;
@@ -67,10 +68,15 @@ public class SamFileValidator {
     private Set<Type> errorsToIgnore = EnumSet.noneOf(Type.class);
     private boolean ignoreWarnings = false;
     private boolean bisulfiteSequenced = false;
+    private boolean validateIndex = false;
     private boolean sequenceDictionaryEmptyAndNoWarningEmitted = false;
+    private final int maxTempFiles;
 
-    public SamFileValidator(final PrintWriter out) {
+    private final static Log log = Log.getInstance(SamFileValidator.class);
+
+    public SamFileValidator(final PrintWriter out, final int maxTempFiles) {
         this.out = out;
+        this.maxTempFiles = maxTempFiles;
     }
 
     /** Sets one or more error types that should not be reported on. */
@@ -92,7 +98,7 @@ public class SamFileValidator {
      * @return boolean  true if there are no validation errors, otherwise false
      */
     public boolean validateSamFileSummary(final SAMFileReader samReader, final ReferenceSequenceFile reference) {
-        init(reference);
+        init(reference, samReader.getFileHeader());
 
         validateSamFile(samReader, out);
 
@@ -123,7 +129,7 @@ public class SamFileValidator {
      * @return boolean  true if there are no validation errors, otherwise false
      */
     public boolean validateSamFileVerbose(final SAMFileReader samReader, final ReferenceSequenceFile reference) {
-        init(reference);
+        init(reference, samReader.getFileHeader());
 
         try {
             validateSamFile(samReader, out);
@@ -168,6 +174,14 @@ public class SamFileValidator {
             validateHeader(samReader.getFileHeader());
             orderChecker = new SAMSortOrderChecker(samReader.getFileHeader().getSortOrder());
             validateSamRecords(samReader);
+            validateUnmatchedPairs();
+            if (validateIndex){
+                try {
+                   BamIndexValidator.exhaustivelyTestIndex(samReader);
+                } catch (Exception e){
+                    addError(new SAMValidationError(Type.INVALID_INDEX_FILE_POINTER, e.getMessage(), null));
+                }
+            }
             
             if (errorsByType.isEmpty()) {
                 out.println("No errors found");
@@ -177,10 +191,47 @@ public class SamFileValidator {
         }
     }
 
+
+    /**
+     * Report on reads marked as paired, for which the mate was not found.
+     */
+    private void validateUnmatchedPairs() {
+        final InMemoryPairEndInfoMap inMemoryPairMap;
+        if (pairEndInfoByName instanceof CoordinateSortedPairEndInfoMap) {
+            // For the coordinate-sorted map, need to detect mate pairs in which the mateReferenceIndex on one end
+            // does not match the readReference index on the other end, so the pairs weren't united and validated.
+            inMemoryPairMap = new InMemoryPairEndInfoMap();
+            CloseableIterator<Map.Entry<String, PairEndInfo>> it = ((CoordinateSortedPairEndInfoMap)pairEndInfoByName).iterator();
+            while (it.hasNext()) {
+                Map.Entry<String, PairEndInfo> entry = it.next();
+                PairEndInfo pei = inMemoryPairMap.remove(entry.getValue().readReferenceIndex, entry.getKey());
+                if (pei != null) {
+                    // Found a mismatch btw read.mateReferenceIndex and mate.readReferenceIndex
+                    List<SAMValidationError> errors = pei.validateMates(entry.getValue(), entry.getKey());
+                    for (final SAMValidationError error : errors) {
+                        addError(error);
+                    }
+                } else {
+                    // Mate not found.
+                    inMemoryPairMap.put(entry.getValue().mateReferenceIndex, entry.getKey(), entry.getValue());
+                }
+            }
+            it.close();
+        } else {
+            inMemoryPairMap = (InMemoryPairEndInfoMap)pairEndInfoByName;
+        }
+        // At this point, everything in InMemoryMap is a read marked as a pair, for which a mate was not found.
+        for (final Map.Entry<String, PairEndInfo> entry: inMemoryPairMap) {
+            addError(new SAMValidationError(Type.MATE_NOT_FOUND, "Mate not found for paired read", entry.getKey()));
+        }
+    }
+
     private void validateSamRecords(final Iterable<SAMRecord> samRecords) {
         long recordNumber = 1;
+        SAMRecordIterator iter = (SAMRecordIterator) samRecords.iterator();
         try {
-            for (final SAMRecord record : samRecords) {
+            while(iter.hasNext()){
+                SAMRecord record = iter.next();
                 final Collection<SAMValidationError> errors = record.isValid();
                 if (errors != null) {
                     for (final SAMValidationError error : errors) {
@@ -203,6 +254,9 @@ public class SamFileValidator {
 
                 }
                 recordNumber++;
+                if (recordNumber % 10000000 == 0) {
+                    log.info(recordNumber + " reads validated.");
+                }
             }
         } catch (SAMFormatException e) {
             // increment record number because the iterator behind the SAMFileReader
@@ -211,6 +265,8 @@ public class SamFileValidator {
             throw new PicardException("SAMFormatException on record " + recordNumber, e);
         } catch (FileTruncatedException e) {
             addError(new SAMValidationError(Type.TRUNCATED_FILE, "File is truncated", null));
+        } finally {
+            iter.close();
         }
     }
 
@@ -293,8 +349,12 @@ public class SamFileValidator {
         }
     }
     
-    private void init(final ReferenceSequenceFile reference) {
-        this.pairEndInfoByName = new HashMap<String, PairEndInfo>();
+    private void init(final ReferenceSequenceFile reference, final SAMFileHeader header) {
+        if (header.getSortOrder() == SAMFileHeader.SortOrder.coordinate) {
+            this.pairEndInfoByName = new CoordinateSortedPairEndInfoMap();
+        } else {
+            this.pairEndInfoByName = new InMemoryPairEndInfoMap();
+        }
         if (reference != null) {
             this.refFileWalker = new ReferenceSequenceFileWalker(reference);
         }
@@ -337,9 +397,9 @@ public class SamFileValidator {
             return;
         }
         
-        final PairEndInfo pairEndInfo = pairEndInfoByName.remove(record.getReadName());
+        final PairEndInfo pairEndInfo = pairEndInfoByName.remove(record.getReferenceIndex(), record.getReadName());
         if (pairEndInfo == null) {
-            pairEndInfoByName.put(record.getReadName(), new PairEndInfo(record, recordNumber));
+            pairEndInfoByName.put(record.getMateReferenceIndex(), record.getReadName(), new PairEndInfo(record, recordNumber));
         } else {
             final List<SAMValidationError> errors =
                 pairEndInfo.validateMates(new PairEndInfo(record, recordNumber), record.getReadName());
@@ -355,9 +415,10 @@ public class SamFileValidator {
         }
         if (fileHeader.getVersion() == null) {
             addError(new SAMValidationError(Type.MISSING_VERSION_NUMBER, "Header has no version number", null));
-        } else if (!fileHeader.getVersion().equals(SAMFileHeader.CURRENT_VERSION)) {
+        } else if (!SAMFileHeader.ACCEPTABLE_VERSIONS.contains(fileHeader.getVersion())) {
             addError(new SAMValidationError(Type.INVALID_VERSION_NUMBER, "Header version: " +
-                    fileHeader.getVersion() + " does not match expected version: " + SAMFileHeader.CURRENT_VERSION,
+                    fileHeader.getVersion() + " does not match any of the acceptable versions: " +
+                    StringUtil.join(", ", SAMFileHeader.ACCEPTABLE_VERSIONS.toArray(new String[0])),
                     null));
         }
         if (fileHeader.getSequenceDictionary().isEmpty()) {
@@ -366,6 +427,26 @@ public class SamFileValidator {
         if (fileHeader.getReadGroups().isEmpty()) {
             addError(new SAMValidationError(Type.MISSING_READ_GROUP, "Read groups is empty", null));
         }
+        List<SAMProgramRecord> pgs = fileHeader.getProgramRecords();
+        for (int i = 0; i < pgs.size()-1; i++) {
+            for (int j=i+1; j < pgs.size(); j++) {
+                if (pgs.get(i).getProgramGroupId().equals(pgs.get(j).getProgramGroupId())) {
+                    addError(new SAMValidationError(Type.DUPLICATE_PROGRAM_GROUP_ID, "Duplicate " +
+                            "program group id: " + pgs.get(i).getProgramGroupId(), null));
+                }
+            }
+        }
+
+        List<SAMReadGroupRecord> rgs = fileHeader.getReadGroups();
+        for (int i = 0; i < rgs.size()-1; i++) {
+            for (int j=i+1; j < rgs.size(); j++) {
+                if (rgs.get(i).getReadGroupId().equals(rgs.get(j).getReadGroupId())) {
+                    addError(new SAMValidationError(Type.DUPLICATE_READ_GROUP_ID, "Duplicate " +
+                            "read group id: " + rgs.get(i).getReadGroupId(), null));
+                }
+            }
+        }
+
     }
 
     private void addError(final SAMValidationError error) {
@@ -397,6 +478,13 @@ public class SamFileValidator {
     public boolean isBisulfiteSequenced() { return bisulfiteSequenced; }
     public void setBisulfiteSequenced(boolean bisulfiteSequenced) { this.bisulfiteSequenced = bisulfiteSequenced; }
 
+    public SamFileValidator setValidateIndex(boolean validateIndex) {
+        // The SAMFileReader must also have IndexCaching enabled to have the index validated,
+        // samReader.enableIndexCaching(true);
+        this.validateIndex = validateIndex;
+        return this;
+    }
+
     public static class ValidationMetrics extends MetricBase {
     }
     
@@ -430,7 +518,19 @@ public class SamFileValidator {
             this.mateReferenceIndex = record.getMateReferenceIndex();
             this.mateUnmappedFlag = record.getMateUnmappedFlag();
         }
-        
+
+        private PairEndInfo(int readAlignmentStart, int readReferenceIndex, boolean readNegStrandFlag, boolean readUnmappedFlag, int mateAlignmentStart, int mateReferenceIndex, boolean mateNegStrandFlag, boolean mateUnmappedFlag, long recordNumber) {
+            this.readAlignmentStart = readAlignmentStart;
+            this.readReferenceIndex = readReferenceIndex;
+            this.readNegStrandFlag = readNegStrandFlag;
+            this.readUnmappedFlag = readUnmappedFlag;
+            this.mateAlignmentStart = mateAlignmentStart;
+            this.mateReferenceIndex = mateReferenceIndex;
+            this.mateNegStrandFlag = mateNegStrandFlag;
+            this.mateUnmappedFlag = mateUnmappedFlag;
+            this.recordNumber = recordNumber;
+        }
+
         public List<SAMValidationError> validateMates(final PairEndInfo mate, final String readName) {
             final List<SAMValidationError> errors = new ArrayList<SAMValidationError>();
             validateMateFields(this, mate, readName, errors);
@@ -476,4 +576,108 @@ public class SamFileValidator {
             super("maxVerboseOutput exceeded.");
         }
     }
+
+    interface PairEndInfoMap extends Iterable<Map.Entry<String, PairEndInfo>> {
+        void put(int mateReferenceIndex, String key, PairEndInfo value);
+        PairEndInfo remove(int mateReferenceIndex, String key);
+        CloseableIterator<Map.Entry<String, PairEndInfo>> iterator();
+    }
+
+    private class CoordinateSortedPairEndInfoMap implements PairEndInfoMap {
+        private final CoordinateSortedPairInfoMap<String, PairEndInfo> onDiskMap =
+                new CoordinateSortedPairInfoMap<String, PairEndInfo>(maxTempFiles, new Codec());
+
+        public void put(int mateReferenceIndex, String key, PairEndInfo value) {
+            onDiskMap.put(mateReferenceIndex, key, value);
+        }
+
+        public PairEndInfo remove(int mateReferenceIndex, String key) {
+            return onDiskMap.remove(mateReferenceIndex, key); 
+        }
+
+        public CloseableIterator<Map.Entry<String, PairEndInfo>> iterator() {
+            return onDiskMap.iterator();
+        }
+
+        private class Codec implements CoordinateSortedPairInfoMap.Codec<String, PairEndInfo> {
+            private DataInputStream in;
+            private DataOutputStream out;
+
+            public void setOutputStream(final OutputStream os) { this.out = new DataOutputStream(os); }
+            public void setInputStream(final InputStream is) { this.in = new DataInputStream(is); }
+
+            public void encode(String key, PairEndInfo record) {
+                try {
+                    out.writeUTF(key);
+                    out.writeInt(record.readAlignmentStart);
+                    out.writeInt(record.readReferenceIndex);
+                    out.writeBoolean(record.readNegStrandFlag);
+                    out.writeBoolean(record.readUnmappedFlag);
+                    out.writeInt(record.mateAlignmentStart);
+                    out.writeInt(record.mateReferenceIndex);
+                    out.writeBoolean(record.mateNegStrandFlag);
+                    out.writeBoolean(record.mateUnmappedFlag);
+                    out.writeLong(record.recordNumber);
+                } catch (IOException e) {
+                    throw new PicardException("Error spilling PairInfo to disk", e);
+                }
+            }
+
+            public Map.Entry<String, PairEndInfo> decode() {
+                try {
+                    final String key = in.readUTF();
+                    final int readAlignmentStart = in.readInt();
+                    final int readReferenceIndex = in.readInt();
+                    final boolean readNegStrandFlag = in.readBoolean();
+                    final boolean readUnmappedFlag = in.readBoolean();
+
+                    final int mateAlignmentStart = in.readInt();
+                    final int mateReferenceIndex = in.readInt();
+                    final boolean mateNegStrandFlag = in.readBoolean();
+                    final boolean mateUnmappedFlag = in.readBoolean();
+
+                    final long recordNumber = in.readLong();
+                    final PairEndInfo rec = new PairEndInfo(readAlignmentStart, readReferenceIndex, readNegStrandFlag, readUnmappedFlag, mateAlignmentStart, mateReferenceIndex, mateNegStrandFlag, mateUnmappedFlag, recordNumber);
+                    return new AbstractMap.SimpleEntry(key, rec);
+                } catch (IOException e) {
+                    throw new PicardException("Error reading PairInfo from disk", e);
+                }
+            }
+        }
+    }
+
+    private static class InMemoryPairEndInfoMap implements PairEndInfoMap {
+        private final Map<String, PairEndInfo> map = new HashMap<String, PairEndInfo>();
+
+        public void put(int mateReferenceIndex, String key, PairEndInfo value) {
+            if (mateReferenceIndex != value.mateReferenceIndex)
+                throw new IllegalArgumentException("mateReferenceIndex does not agree with PairEndInfo");
+            map.put(key, value);
+        }
+
+        public PairEndInfo remove(int mateReferenceIndex, String key) {
+            return map.remove(key);
+        }
+
+        public CloseableIterator<Map.Entry<String, PairEndInfo>> iterator() {
+            final Iterator<Map.Entry<String, PairEndInfo>> it = map.entrySet().iterator();
+            return new CloseableIterator<Map.Entry<String, PairEndInfo>>() {
+                public void close() {
+                    // do nothing
+                }
+
+                public boolean hasNext() {
+                    return it.hasNext();
+                }
+
+                public Map.Entry<String, PairEndInfo> next() {
+                    return it.next();
+                }
+
+                public void remove() {
+                    it.remove();
+                }
+            };
+        }
+    }
 }
