@@ -24,7 +24,6 @@
 package net.sf.samtools;
 
 import net.sf.samtools.util.RuntimeIOException;
-import static net.sf.samtools.util.BlockCompressedInputStream.getFileBlock;
 
 import java.io.*;
 import java.nio.channels.FileChannel;
@@ -43,35 +42,61 @@ import java.util.*;
  * Anyone wanting to implement a reader for a differing
  * or extended BAM index format should implement BAMIndex directly.
  */
-abstract class AbstractBAMFileIndex implements BAMIndex {
+public abstract class AbstractBAMFileIndex implements BAMIndex {
 
     /**
      * Reports the total amount of genomic data that any bin can index.
      */
-    protected static final int BIN_SPAN = 512*1024*1024;
+    protected static final int BIN_GENOMIC_SPAN = 512*1024*1024;
 
     /**
      * What is the starting bin for each level?
      */
     private static final int[] LEVEL_STARTS = {0,1,9,73,585,4681};
 
+    /**
+     * Reports the maximum number of bins that can appear in a BAM file.
+     */
+    public static final int MAX_BINS = 37450;   // =(8^6-1)/7+1
+    
     public static final int MAX_LINEAR_INDEX_SIZE = MAX_BINS+1-LEVEL_STARTS[LEVEL_STARTS.length-1];
 
     private final File mFile;
-    private MappedByteBuffer mFileBuffer;
+    private final IndexFileBuffer mIndexBuffer;
 
-    protected AbstractBAMFileIndex(final File file) {
+    private SAMSequenceDictionary mBamDictionary = null;
+
+    protected AbstractBAMFileIndex(final File file, final SAMSequenceDictionary dictionary) {
+        this(file, dictionary, true);
+    }
+
+    protected AbstractBAMFileIndex(final File file, final SAMSequenceDictionary dictionary, boolean useMemoryMapping) {
         mFile = file;
-        if (file != null){
-            open();
+        mBamDictionary = dictionary;
+        mIndexBuffer = (useMemoryMapping ? new MemoryMappedFileBuffer(file) : new RandomAccessFileBuffer(file));
+
+        // Verify the magic number.
+        seek(0);
+        final byte[] buffer = new byte[4];
+        readBytes(buffer);
+        if (!Arrays.equals(buffer, BAMFileConstants.BAM_INDEX_MAGIC)) {
+            throw new RuntimeException("Invalid file header in BAM index " + mFile +
+                                       ": " + new String(buffer));
         }
     }
 
+    /**
+     * Close this index and release any associated resources.
+     */
+    public void close() {
+        mIndexBuffer.close();
+    }
+
     /**
      * Get the number of levels employed by this index.
      * @return Number of levels in this index.
      */
-    public int getNumIndexLevels() {
+    public static int getNumIndexLevels() {
         return LEVEL_STARTS.length;
     }
 
@@ -80,7 +105,7 @@ abstract class AbstractBAMFileIndex implements BAMIndex {
      * @param levelNumber Level number.  0-based.
      * @return The first bin in this level.
      */
-    public int getFirstBinInLevel(final int levelNumber) {
+    public static int getFirstBinInLevel(final int levelNumber) {
         return LEVEL_STARTS[levelNumber];
     }
 
@@ -120,7 +145,7 @@ abstract class AbstractBAMFileIndex implements BAMIndex {
         final int level = getLevelForBin(bin);
         final int levelStart = LEVEL_STARTS[level];
         final int levelSize = ((level==getNumIndexLevels()-1) ? MAX_BINS-1 : LEVEL_STARTS[level+1]) - levelStart;
-        return (bin.getBinNumber() - levelStart)*(BIN_SPAN/levelSize)+1;
+        return (bin.getBinNumber() - levelStart)*(BIN_GENOMIC_SPAN /levelSize)+1;
     }
 
     /**
@@ -132,57 +157,20 @@ abstract class AbstractBAMFileIndex implements BAMIndex {
         final int level = getLevelForBin(bin);
         final int levelStart = LEVEL_STARTS[level];
         final int levelSize = ((level==getNumIndexLevels()-1) ? MAX_BINS-1 : LEVEL_STARTS[level+1]) - levelStart;
-        return (bin.getBinNumber()-levelStart+1)*(BIN_SPAN/levelSize);
-    }
-
-    public void open() {
-        // Open the file stream.
-        if (mFileBuffer != null) {
-            return;
-        }
-        try {
-            FileInputStream fileStream = new FileInputStream(mFile);
-            FileChannel fileChannel = fileStream.getChannel();
-            mFileBuffer = fileChannel.map(FileChannel.MapMode.READ_ONLY, 0L, fileChannel.size());
-            mFileBuffer.order(ByteOrder.LITTLE_ENDIAN);
-
-            fileChannel.close();
-            fileStream.close();
-        }
-        catch (IOException exc) {
-            throw new RuntimeIOException(exc.getMessage(), exc);
-        }
-
-        // Verify the magic number.
-        seek(0);
-        final byte[] buffer = new byte[4];
-        readBytes(buffer);
-        if (!Arrays.equals(buffer, BAMFileConstants.BAM_INDEX_MAGIC)) {
-            close();
-            throw new RuntimeException("Invalid file header in BAM index " + mFile +
-                                       ": " + new String(buffer));
-        }
+        return (bin.getBinNumber()-levelStart+1)*(BIN_GENOMIC_SPAN /levelSize);
     }
 
     public int getNumberOfReferences() {
-        if(mFileBuffer == null)
-            throw new SAMException("Cannot query a closed index file");
         seek(4);
         return readInteger();
     }
 
-    public void close() {
-        mFileBuffer = null;
-    }
-
     /**
      * Use to get close to the unmapped reads at the end of a BAM file.
      * @return The file offset of the first record in the last linear bin, or -1
      * if there are no elements in linear bins (i.e. no mapped reads).
      */
     public long getStartOfLastLinearBin() {
-        if(mFileBuffer == null)
-            throw new SAMException("Cannot query a closed index file");
         seek(4);
 
         final int sequenceCount = readInteger();
@@ -212,27 +200,68 @@ abstract class AbstractBAMFileIndex implements BAMIndex {
         return lastLinearIndexPointer;
     }
 
-     /**
-     * Can only be called once all other references have been read, before the file is closed
-     * @return meta data at the end of the bam index that indicates count of records holding no coordinates
+    /**
+     * Return meta data for the given reference including information about number of aligned, unaligned, and noCoordinate records
+     *
+     * @param reference the reference of interest
+     * @return meta data for the reference
      */
-    Long getNoCoordinateCount(){
-       if(mFileBuffer == null)
-            throw new SAMException("Cannot query a closed index file");
+    public BAMIndexMetaData getMetaData(int reference) {
+        seek(4);
+
+        List<Chunk> metaDataChunks = new ArrayList<Chunk>();
+
+        final int sequenceCount = readInteger();
+
+        if (reference >= sequenceCount) {
+            return null;
+        }
+
+        skipToSequence(reference);
+
+        int binCount = readInteger();
+        for (int binNumber = 0; binNumber < binCount; binNumber++) {
+            final int indexBin = readInteger();
+            final int nChunks = readInteger();
+            // System.out.println("# bin[" + i + "] = " + indexBin + ", nChunks = " + nChunks);
+            Chunk lastChunk = null;
+            if (indexBin == MAX_BINS) {
+                for (int ci = 0; ci < nChunks; ci++) {
+                    final long chunkBegin = readLong();
+                    final long chunkEnd = readLong();
+                    lastChunk = new Chunk(chunkBegin, chunkEnd);
+                    metaDataChunks.add(lastChunk);
+                }
+                continue;
+            } else {
+                skipBytes(16 * nChunks);
+            }
+        }
+        return new BAMIndexMetaData(metaDataChunks);
+    }
+
+    /**
+     * Returns count of records unassociated with any reference. Call before the index file is closed
+     *
+     * @return meta data at the end of the bam index that indicates count of records holding no coordinates
+     * or null if no meta data (old index format)
+     */
+    public Long getNoCoordinateCount() {
+
+        seek(4);
+        final int sequenceCount = readInteger();
+
+        skipToSequence(sequenceCount);
         try { // in case of old index file without meta data
-           return readLong();
-        } catch (Exception e){
+            return readLong();
+        } catch (Exception e) {
             return null;
         }
     }
 
     protected BAMIndexContent query(final int referenceSequence, final int startPos, final int endPos) {
-        if(mFileBuffer == null)
-            throw new SAMException("Cannot query a closed index file");
         seek(4);
 
-        List<Bin> bins = null;
-        LinearIndex linearIndex = null;
         List<Chunk> metaDataChunks = new ArrayList<Chunk>();
 
         final int sequenceCount = readInteger();
@@ -248,12 +277,13 @@ abstract class AbstractBAMFileIndex implements BAMIndex {
 
         skipToSequence(referenceSequence);
 
-        final int binCount = readInteger();
-        bins = new ArrayList<Bin>(binCount);
+        int binCount = readInteger();
+        boolean metaDataSeen = false;
+        Bin[] bins = new Bin[getMaxBinNumberForReference(referenceSequence) +1];
         for (int binNumber = 0; binNumber < binCount; binNumber++) {
-            List<Chunk> chunks = new ArrayList<Chunk>();
             final int indexBin = readInteger();
             final int nChunks = readInteger();
+            List<Chunk> chunks = new ArrayList<Chunk>(nChunks);
             // System.out.println("# bin[" + i + "] = " + indexBin + ", nChunks = " + nChunks);
             Chunk lastChunk = null;
             if (regionBins.get(indexBin)) {
@@ -272,21 +302,21 @@ abstract class AbstractBAMFileIndex implements BAMIndex {
                     lastChunk = new Chunk(chunkBegin, chunkEnd);
                     metaDataChunks.add(lastChunk);
                 }
+                metaDataSeen = true;
+                continue; // don't create a Bin
             } else {
                 skipBytes(16 * nChunks);
             }
             Bin bin = new Bin(referenceSequence, indexBin);
             bin.setChunkList(chunks);
             bin.setLastChunk(lastChunk);
-            bins.add(bin);
+            bins[indexBin] = bin;
         }
-        // Reorder the bins in binNumber order.
-        Collections.sort(bins);
 
         final int nLinearBins = readInteger();
 
         final int regionLinearBinStart = LinearIndex.convertToLinearIndexOffset(startPos);
-        final int regionLinearBinStop = LinearIndex.convertToLinearIndexOffset(endPos)>0 ? LinearIndex.convertToLinearIndexOffset(endPos) : nLinearBins-1;
+        final int regionLinearBinStop = endPos > 0 ? LinearIndex.convertToLinearIndexOffset(endPos) : nLinearBins-1;
         final int actualStop = Math.min(regionLinearBinStop, nLinearBins -1);
 
         long[] linearIndexEntries = new long[0];
@@ -297,9 +327,31 @@ abstract class AbstractBAMFileIndex implements BAMIndex {
                 linearIndexEntries[linearBin-regionLinearBinStart] = readLong();
         }
 
-        linearIndex = new LinearIndex(referenceSequence,regionLinearBinStart,linearIndexEntries);
+        final LinearIndex linearIndex = new LinearIndex(referenceSequence,regionLinearBinStart,linearIndexEntries);
 
-        return new BAMIndexContent(referenceSequence, bins, metaDataChunks, linearIndex);
+        return new BAMIndexContent(referenceSequence, bins, binCount - (metaDataSeen? 1 : 0), new BAMIndexMetaData(metaDataChunks), linearIndex);
+    }
+
+    /**
+     * The maximum possible bin number for this reference sequence.
+     * This is based on the maximum coordinate position of the reference
+     * which is based on the size of the reference
+     */
+    private int getMaxBinNumberForReference(final int reference) {
+        try {
+            final int sequenceLength = mBamDictionary.getSequence(reference).getSequenceLength();
+            return getMaxBinNumberForSequenceLength(sequenceLength);
+        } catch (Exception e) {
+            return MAX_BINS;
+        }
+    }
+
+    /**
+     * The maxiumum bin number for a reference sequence of a given length
+     */
+    static int getMaxBinNumberForSequenceLength(int sequenceLength) {
+        return getFirstBinInLevel(getNumIndexLevels() - 1) + (sequenceLength >> 14);
+        // return 4680 + (sequenceLength >> 14); // note 4680 = getFirstBinInLevel(getNumIndexLevels() - 1)
     }
 
     abstract protected BAMIndexContent getQueryResults(int reference);
@@ -309,7 +361,7 @@ abstract class AbstractBAMFileIndex implements BAMIndex {
      * @return How many bins could possibly be used according to this indexing scheme to index a single contig.
      */
     protected int getMaxAddressibleGenomicLocation() {
-        return BIN_SPAN;
+        return BIN_GENOMIC_SPAN;
     }
 
     /**
@@ -342,7 +394,7 @@ abstract class AbstractBAMFileIndex implements BAMIndex {
         final List<Chunk> result = new ArrayList<Chunk>();
         for (final Chunk chunk : chunks) {
             if (chunk.getChunkEnd() <= minimumOffset) {
-                continue;
+                continue;               // linear index optimization
             }
             if (result.isEmpty()) {
                 result.add(chunk);
@@ -351,9 +403,7 @@ abstract class AbstractBAMFileIndex implements BAMIndex {
             }
             // Coalesce chunks that are in adjacent file blocks.
             // This is a performance optimization.
-            final long lastFileBlock = getFileBlock(lastChunk.getChunkEnd());
-            final long chunkFileBlock = getFileBlock(chunk.getChunkStart());
-            if (chunkFileBlock - lastFileBlock > 1) {
+            if (!lastChunk.overlaps(chunk) && !lastChunk.isAdjacentTo(chunk)) {
                 result.add(chunk);
                 lastChunk = chunk;
             } else {
@@ -383,22 +433,187 @@ abstract class AbstractBAMFileIndex implements BAMIndex {
     }
 
     private void readBytes(final byte[] bytes) {
-        mFileBuffer.get(bytes);
+        mIndexBuffer.readBytes(bytes);
     }
 
     private int readInteger() {
-        return mFileBuffer.getInt();
+        return mIndexBuffer.readInteger();
     }
 
     private long readLong() {
-        return mFileBuffer.getLong();
+        return mIndexBuffer.readLong();
     }
 
     private void skipBytes(final int count) {
-        mFileBuffer.position(mFileBuffer.position() + count);
+        mIndexBuffer.skipBytes(count);
     }
 
     private void seek(final int position) {
-        mFileBuffer.position(position);
+        mIndexBuffer.seek(position);
     }
-}
\ No newline at end of file
+
+    private abstract static class IndexFileBuffer {
+        abstract void readBytes(final byte[] bytes);
+        abstract int readInteger();
+        abstract long readLong();
+        abstract void skipBytes(final int count);
+        abstract void seek(final int position);
+        abstract void close();
+    }
+
+    /**
+     * Traditional implementation of BAM index file access using memory mapped files.
+     */
+    private static class MemoryMappedFileBuffer extends IndexFileBuffer {
+        private MappedByteBuffer mFileBuffer;
+
+        MemoryMappedFileBuffer(File file) {
+            try {
+                // Open the file stream.
+                FileInputStream fileStream = new FileInputStream(file);
+                FileChannel fileChannel = fileStream.getChannel();
+                mFileBuffer = fileChannel.map(FileChannel.MapMode.READ_ONLY, 0L, fileChannel.size());
+                mFileBuffer.order(ByteOrder.LITTLE_ENDIAN);
+                fileChannel.close();
+                fileStream.close();
+            } catch (IOException exc) {
+                throw new RuntimeIOException(exc.getMessage(), exc);
+            }
+        }
+
+        void readBytes(final byte[] bytes) {
+            mFileBuffer.get(bytes);
+        }
+
+        int readInteger() {
+            return mFileBuffer.getInt();
+        }
+
+        long readLong() {
+            return mFileBuffer.getLong();
+        }
+
+        void skipBytes(final int count) {
+            mFileBuffer.position(mFileBuffer.position() + count);
+        }
+
+        void seek(final int position) {
+            mFileBuffer.position(position);
+        }
+
+        void close() {
+            mFileBuffer = null;
+        }
+    }
+
+    /**
+     * Alternative implementation of BAM index file access using regular I/O instead of memory mapping.
+     * 
+     * This implementation can be more scalable for certain applications that need to access large numbers of BAM files.
+     * Java provides no way to explicitly release a memory mapping.  Instead, you need to wait for the garbage collector
+     * to finalize the MappedByteBuffer.  Because of this, when accessing many BAM files or when querying many BAM files
+     * sequentially, you cannot easily control the physical memory footprint of the java process.
+     * This can limit scalability and can have bad interactions with load management software like LSF, forcing you
+     * to reserve enough physical memory for a worst case scenario.
+     * The use of regular I/O allows you to trade somewhat slower performance for a small, fixed memory footprint
+     * if that is more suitable for your application.
+     */
+    private static class RandomAccessFileBuffer extends IndexFileBuffer {
+        private static final int PAGE_SIZE = 4 * 1024;
+        private static final int PAGE_OFFSET_MASK = PAGE_SIZE-1;
+        private static final int PAGE_MASK = ~PAGE_OFFSET_MASK;
+        private static final int INVALID_PAGE = 1;
+        private File mFile;
+        private RandomAccessFile mRandomAccessFile;
+        private int mFileLength;
+        private int mFilePointer = 0;
+        private int mCurrentPage = INVALID_PAGE;
+        private final byte[] mBuffer = new byte[PAGE_SIZE];
+
+        RandomAccessFileBuffer(File file) {
+            mFile = file;
+            try {
+                mRandomAccessFile = new RandomAccessFile(file, "r");
+                long fileLength = mRandomAccessFile.length();
+                if (fileLength > Integer.MAX_VALUE) {
+                    throw new RuntimeException("BAM index file " + mFile + " is too large: " + fileLength);
+                }
+                mFileLength = (int) fileLength;
+            } catch (IOException exc) {
+                throw new RuntimeIOException(exc.getMessage(), exc);
+            }
+        }
+
+        void readBytes(final byte[] bytes) {
+            int resultOffset = 0;
+            int resultLength = bytes.length;
+            if (mFilePointer + resultLength > mFileLength) {
+                throw new RuntimeException("Attempt to read past end of BAM index file (file is truncated?): " + mFile);
+            }
+            while (resultLength > 0) {
+                loadPage(mFilePointer);
+                final int pageOffset = mFilePointer & PAGE_OFFSET_MASK;
+                final int copyLength = Math.min(resultLength, PAGE_SIZE - pageOffset);
+                System.arraycopy(mBuffer, pageOffset, bytes, resultOffset, copyLength);
+                mFilePointer += copyLength;
+                resultOffset += copyLength;
+                resultLength -= copyLength;
+            }
+        }
+
+        int readInteger() {
+            // This takes advantage of the fact that integers in BAM index files are always 4-byte aligned.
+            loadPage(mFilePointer);
+            final int pageOffset = mFilePointer & PAGE_OFFSET_MASK;
+            mFilePointer += 4;
+            return((mBuffer[pageOffset + 0] & 0xFF) |
+                   ((mBuffer[pageOffset + 1] & 0xFF) << 8) | 
+                   ((mBuffer[pageOffset + 2] & 0xFF) << 16) |
+                   ((mBuffer[pageOffset + 3] & 0xFF) << 24));
+        }
+
+        long readLong() {
+            // BAM index files are always 4-byte aligned, but not necessrily 8-byte aligned.
+            // So, rather than fooling with complex page logic we simply read the long in two 4-byte chunks.
+            long lower = readInteger();
+            long upper = readInteger();
+            return ((upper << 32) | (lower & 0xFFFFFFFFL));
+        }
+
+        void skipBytes(final int count) {
+            mFilePointer += count;
+        }
+        
+        void seek(final int position) {
+            mFilePointer = position;
+        }
+
+        void close() {
+            mFilePointer = 0;
+            mCurrentPage = INVALID_PAGE;
+            if (mRandomAccessFile != null) {
+                try {
+                    mRandomAccessFile.close();
+                } catch (IOException exc) {
+                    throw new RuntimeIOException(exc.getMessage(), exc);
+                }
+                mRandomAccessFile = null;
+            }
+        }
+
+        private void loadPage(int filePosition) {
+            final int page = filePosition & PAGE_MASK;
+            if (page == mCurrentPage) {
+                return;
+            }
+            try {
+                mRandomAccessFile.seek(page);
+                final int readLength = Math.min(mFileLength - page, PAGE_SIZE);
+                mRandomAccessFile.readFully(mBuffer, 0, readLength);
+                mCurrentPage = page;
+            } catch (IOException exc) {
+                throw new RuntimeIOException("Exception reading BAM index file " + mFile + ": " + exc.getMessage(), exc);
+            }
+        }
+    }
+}
