@@ -56,7 +56,7 @@ public class SequenceUtil {
 
         return lhs == rhs;
     }
-    
+
     /**
      * returns true if the value of base represents a no call
      */
@@ -64,6 +64,14 @@ public class SequenceUtil {
         return base == 'N' || base == 'n' || base == '.';
     }
 
+    /** Returns true if the byte is in [acgtACGT]. */
+    public static boolean isValidBase(final byte b){
+        return b == a || b == A ||
+               b == c || b == C ||
+               b == g || b == G ||
+               b == t || b == T;
+    }
+
     /**
      * Throws an exception if both parameters are not null sequenceListsEqual returns false
      * @param s1 a list of sequence headers
@@ -79,11 +87,11 @@ public class SequenceUtil {
         for (int i = 0; i < s1.size(); ++i) {
             if (!s1.get(i).isSameSequence(s2.get(i))) {
                 String s1Attrs = "";
-                for (final java.util.Map.Entry<String, Object> entry : s1.get(i).getAttributes()) {
+                for (final java.util.Map.Entry<String, String> entry : s1.get(i).getAttributes()) {
                     s1Attrs += "/" + entry.getKey() + "=" + entry.getValue();
                 }
                 String s2Attrs = "";
-                for (final java.util.Map.Entry<String, Object> entry : s2.get(i).getAttributes()) {
+                for (final java.util.Map.Entry<String, String> entry : s2.get(i).getAttributes()) {
                     s2Attrs += "/" + entry.getKey() + "=" + entry.getValue();
                 }
                 throw new SequenceListsDifferException("Sequences at index " + i + " don't match: " +
@@ -213,30 +221,34 @@ public class SequenceUtil {
      */
     public static int countMismatches(final SAMRecord read, final byte[] referenceBases, final int referenceOffset,
                                       final boolean bisulfiteSequence) {
-        int mismatches = 0;
+        try {
+            int mismatches = 0;
 
-        final byte[] readBases = read.getReadBases();
+            final byte[] readBases = read.getReadBases();
 
-        for (final AlignmentBlock block : read.getAlignmentBlocks()) {
-            final int readBlockStart = block.getReadStart() - 1;
-            final int referenceBlockStart = block.getReferenceStart() - 1 - referenceOffset;
-            final int length = block.getLength();
+            for (final AlignmentBlock block : read.getAlignmentBlocks()) {
+                final int readBlockStart = block.getReadStart() - 1;
+                final int referenceBlockStart = block.getReferenceStart() - 1 - referenceOffset;
+                final int length = block.getLength();
 
-            for (int i=0; i<length; ++i) {
-                if (!bisulfiteSequence) {
-                    if (!basesEqual(readBases[readBlockStart+i], referenceBases[referenceBlockStart+i])) {
-                        ++mismatches;
+                for (int i=0; i<length; ++i) {
+                    if (!bisulfiteSequence) {
+                        if (!basesEqual(readBases[readBlockStart+i], referenceBases[referenceBlockStart+i])) {
+                            ++mismatches;
+                        }
                     }
-                }
-                else {
-                    if (!bisulfiteBasesEqual(read.getReadNegativeStrandFlag(), readBases[readBlockStart+i],
-                            referenceBases[referenceBlockStart+i])) {
-                        ++mismatches;
+                    else {
+                        if (!bisulfiteBasesEqual(read.getReadNegativeStrandFlag(), readBases[readBlockStart+i],
+                                referenceBases[referenceBlockStart+i])) {
+                            ++mismatches;
+                        }
                     }
                 }
             }
+            return mismatches;
+        } catch (Exception e) {
+            throw new SAMException("Exception counting mismatches for read " + read, e);
         }
-        return mismatches;
     }
 
     /**
@@ -364,6 +376,30 @@ public class SequenceUtil {
         return qualities;
     }
 
+    public static int countInsertedBases(final Cigar cigar) {
+        int ret = 0;
+        for (final CigarElement element : cigar.getCigarElements()) {
+            if (element.getOperator() == CigarOperator.INSERTION) ret += element.getLength();
+        }
+        return ret;
+    }
+
+    public static int countDeletedBases(final Cigar cigar) {
+        int ret = 0;
+        for (final CigarElement element : cigar.getCigarElements()) {
+            if (element.getOperator() == CigarOperator.DELETION) ret += element.getLength();
+        }
+        return ret;
+    }
+
+    public static int countInsertedBases(final SAMRecord read) {
+        return countInsertedBases(read.getCigar());
+    }
+
+    public static int countDeletedBases(final SAMRecord read) {
+        return countDeletedBases(read.getCigar());
+    }
+
     /**
      * Calculates the for the predefined NM tag from the SAM spec. To the result of
      * countMismatches() it adds 1 for each indel.
@@ -383,6 +419,7 @@ public class SequenceUtil {
                                         final int referenceOffset) {
         return calculateSamNmTag(read, referenceBases, referenceOffset, false);
     }
+
     /**
      * Calculates the for the predefined NM tag from the SAM spec. To the result of
      * countMismatches() it adds 1 for each indel.
@@ -451,10 +488,23 @@ public class SequenceUtil {
         }
     }
 
+    
+    /** Reverses the quals in place. */
+    public static void reverseQualities(final byte[] quals) {
+        final int lastIndex = quals.length - 1;
+
+        int i, j;
+        for (i=0, j=lastIndex; i<j; ++i, --j) {
+            final byte tmp = quals[i];
+            quals[i] = quals[j];
+            quals[j] = tmp;
+        }
+    }
+    
     /** Returns true if the bases are equal OR if the mismatch cannot be accounted for by
      * bisfulite treatment.  C->T on the positive strand and G->A on the negative strand
      * do not count as mismatches */
-    public static boolean bisulfiteBasesEqual(boolean negativeStrand, byte read, byte reference) {
+    public static boolean bisulfiteBasesEqual(final boolean negativeStrand, final byte read, final byte reference) {
 
         if (basesEqual(read, reference)) {
             return true;
@@ -496,7 +546,7 @@ public class SequenceUtil {
      * '-'.  If the read is soft-clipped, reference contains '0'.  If there is a skipped region and
      * includeReferenceBasesForDeletions==true, reference will have Ns for the skipped region.
      */
-    public static byte[] makeReferenceFromAlignment(final SAMRecord rec, boolean includeReferenceBasesForDeletions) {
+    public static byte[] makeReferenceFromAlignment(final SAMRecord rec, final boolean includeReferenceBasesForDeletions) {
         final String md = rec.getStringAttribute(SAMTag.MD.name());
         if (md == null) {
             throw new SAMException("Cannot create reference from SAMRecord with no MD tag, read: " + rec.getReadName());
@@ -513,15 +563,15 @@ public class SequenceUtil {
         final byte[] ret = new byte[maxOutputLength];
         int outIndex = 0;
 
-        Matcher match = mdPat.matcher(md);
+        final Matcher match = mdPat.matcher(md);
         int curSeqPos = 0;
 
         int savedBases = 0;
         final byte[] seq = rec.getReadBases();
         for (final CigarElement cigEl : cigar.getCigarElements())
         {
-            int cigElLen = cigEl.getLength();
-            CigarOperator cigElOp = cigEl.getOperator();
+            final int cigElLen = cigEl.getLength();
+            final CigarOperator cigElOp = cigEl.getOperator();
 
 
             if (cigElOp == CigarOperator.SKIPPED_REGION) {
@@ -557,7 +607,7 @@ public class SequenceUtil {
                         if ( ((mg = match.group(1)) !=null) && (mg.length() > 0) )
                         {
                             // It's a number , meaning a series of matches
-                            int num = Integer.parseInt(mg);
+                            final int num = Integer.parseInt(mg);
                             for (int i = 0; i < num; i++)
                             {
                                 if (basesMatched<cigElLen)
@@ -628,7 +678,7 @@ public class SequenceUtil {
                 // We have an insertion in read
                 for (int i = 0; i < cigElLen; i++)
                 {
-                    char c = (cigElOp == CigarOperator.SOFT_CLIP) ? '0' : '-';
+                    final char c = (cigElOp == CigarOperator.SOFT_CLIP) ? '0' : '-';
                     ret[outIndex++] =  StringUtil.charToByte(c);
                     curSeqPos++;
                 }
@@ -640,7 +690,7 @@ public class SequenceUtil {
 
         }
         if (outIndex < ret.length) {
-            byte[] shorter = new byte[outIndex];
+            final byte[] shorter = new byte[outIndex];
             System.arraycopy(ret, 0, shorter, 0, outIndex);
             return shorter;
         }
